from .enums import enum, blood_types, constellations, hands, home_towns

class Idol:
    def __init__(self, char_data: dict):
        self.age = char_data['age']
        self.bday = char_data['birth_day']
        self.bmonth = char_data['birth_month']
        self.btype = enum(blood_types, char_data['blood_type'])
        self.bust = char_data['body_size_1']
        self.waist = char_data['body_size_3']
        self.hip = char_data['body_size_3']
        self.horoscope = enum(constellations, char_data['constellation'])
        self.conventional = char_data['conventional']
        self.favorite = char_data['favorite']
        self.hand = enum(hands, char_data['hand'])
        self.height = char_data['height']
        self.home_town = enum(home_towns, char_data['home_town'])
        self.kana_spaced = char_data['kana_spaced']
        self.kanji_spaced = char_data['kanji_spaced']
        self.name = char_data['name']
        self.name_kana = char_data['name_kana']
        self.personality = char_data['personality']
        self.type = char_data['type']
        self.voice = char_data['voice']
        self.weight = char_data['weight']
        self.icon = char_data['icon_image_ref']


import asyncio
import json
import re
from functools import lru_cache

import aiohttp

from .card import Card
from .enums import enum, rarities
from .idol import Idol
from .infos import Event, Gacha, Info


class KiraraException(Exception):
    def __init__(self, http_status, code, msg):
        self.http_status = http_status
        self.code = code
        self.msg = msg
    
    def __str__(self):
        return f"http status: {self.http_status} {self.code} - {self.msg}"

class Kirara:
    def __init__(self, loop: asyncio.AbstractEventLoop = None, session=True, timeout=50):
        self.prefix = 'https://starlight.kirara.ca/api/v1/'
        self.loop = loop or asyncio.get_event_loop()
        self._autoclose_session = True
        self.retries = timeout

        if isinstance(session, aiohttp.ClientSession):
            self._session = session
        else:
            if session:
                self._session = aiohttp.ClientSession(loop=self.loop)
            else:
                pass

    async def internal_call(self, method, url, payload, params):
        args = dict(params=params)
        args['timeout'] = self.retries
        if not url.startswith('http'):
            url = self.prefix + url

        if payload:
            args['data'] = json.dumps(payload)

        async with self._session.request(method, url, **args) as r:
            r.raise_for_status()
            result = await r.json()
            return result

    @lru_cache(maxsize=None)
    async def get(self, url, args=None, payload=None, params=None, **kwargs):
        reconnect = self.retries

        while reconnect > 0:
            result = await self.internal_call('GET', url, payload, kwargs)
            return result

    @lru_cache(maxsize=None)
    async def post(self, url, args=None, payload=None, params=None, **kwargs):
        reconnect = self.retries

        while reconnect > 0:
            result = await self.internal_call('POST', url, payload, kwargs)
            return result

    async def translate(self, translations: tuple):
        result = await self.post('read_tl', payload=translations)

        return result

    async def get_idol(self, char_id: int):
        data = await self.get(f'char_t/{char_id}')

        return Idol(data['result'][0])

    async def get_card(self, card_id, en_translate=False):
        data = await self.get(f'card_t/{card_id}')

        if data['result'][0] is not None:
            card = Card(data['result'][0])
            if en_translate:
                translations = []
                translations.append(card.title)
                if card.skill is not None:
                    translations.append(card.skill.name)
                if card.lead_skill is not None:
                    translations.append(card.lead_skill.name)
                result = await self.translate(translations)

                for strings, translated in result.items():
                    if translated is None:
                        translated = strings
                card.title = card.title if result.get(
                    card.title) is None else result.get(card.title)
                if card.skill is not None:
                    card.skill.name = card.skill.name if result.get(
                        card.skill.name) is None else result.get(card.skill.name)
                if card.lead_skill is not None:
                    card.lead_skill.name = card.lead_skill.name if result.get(
                        card.lead_skill.name) is None else result.get(card.lead_skill.name)
                
                return card
            else:
                return card
            
    async def get_cards(self, card_ids: list, en_translate=False):
        card_ids = str(card_ids).replace('[', '').replace(']', '')
        card_list = []

        data = await self.get(f"card_t/{card_ids}")

        for card_data in data['result']:
            card_list.append(Card(card_data))

        if en_translate:
            translations = []
            for string in card_list:
                translations.append(string.title)

                if string.skill is not None:
                    translations.append(string.skill.name)
                if string.lead_skill is not None:
                    translations.append(string.lead_skill.name)

            result = await self.translate(tuple(translations))

            for strings, translated in result.items():
                if translated is None:
                    translated = strings
            for card in card_list:
                card.title = card.title if result.get(
                    card.title) is None else result.get(card.title)

                if card.skill is not None:
                    card.skill.name = card.skill.name if result.get(
                        card.skill.name) is None else result.get(card.skill.name)

                if card.lead_skill is not None:
                    card.lead_skill.name = card.lead_skill.name if result.get(
                        card.lead_skill.name) is None else result.get(card.lead_skill.name)

        return card_list

    async def get_image(self, card: 'Card', category):
        categories = {
            'card': card,
            'icon': card.icon,
            'spread': card.spread,
            'sprite': card.sprite,
            'puchi': card.puchi
        }

        if category in categories:
            image = categories.get(category)

            async with await self._session.get(image) as response:

                return await response.read()

    async def get_id(self, category, name, card_rarity=None, position=None, is_title=False):
        cat_list = await self.get(f'list/{category}')
        rarity = enum(rarities, card_rarity)
        card_list = []

        for index, code in enumerate(cat_list['result']):
            if is_title:
                cat_name = cat_list['result'][index]['title']
                if cat_name is None:
                    continue
                else:
                    cat_name = cat_list['result'][index]['title'].lower()
                    match = bool(name[:5].lower() in cat_name[:5])
            else:
                cat_name = cat_list['result'][index]['conventional'].lower()
                match = bool(
                    re.search(r'\b{0}\b'.format(name.lower()), cat_name))

            if category == 'char_t':
                if match:
                    return int(cat_list['result'][index]['chara_id'])
            
            elif category == 'card_t':
                if match:
                    if card_rarity:
                        if int(rarity) == int(cat_list['result'][index]['rarity_dep']['rarity']):
                            card_list.append(
                                int(cat_list['result'][index]['id']))
                    else:
                        card_list.append(int(cat_list['result'][index]['id']))
        if position:
            return card_list[position-1]
        else:
            return card_list
    
    async def get_version(self):
        data = self.get('info')

        return Info(data)
    
    async def get_now(self, category, en_translate=False):
        happening_now = await self.get('happening/now')
        categories = {
            'event': happening_now['events'],
            'gacha': happening_now['gachas']
        }
        happening_list = []

        if category in categories:
            stuff = categories.get(category)

            for index, event in enumerate(stuff):
                if category == 'event':
                    happening_list.append(Event(event))
                else:
                    for gacha, event in enumerate(stuff):
                        happening_list.append(Gacha(event))

        if en_translate:
            for thing in happening_list:
                translations = await self.translate((thing.name,))
                for string, translated in translations.items():
                    if translated is None:
                        thing.name = string
                    else:
                        thing.name = translated

        return happening_list

    async def close(self):
        await self._session.close()



from .enums import enum, attributes
from .skills import Skill, LeadSkill
from .idol import Idol

class Card:
    def __init__(self, card_data: dict):

        self.card_id = card_data['id']
        self.album_id = card_data['album_id']
        self.type = card_data['attribute']
        self.image = card_data['card_image_ref']
        self.has_spread = card_data['has_spread']
        self.chara_id = card_data['chara_id']
        self.chara = Idol(card_data['chara'])
        self.conventional = card_data['chara']['conventional']

        self.spread = card_data['spread_image_ref']
        self.sprite = card_data['sprite_image_ref']
        self.sign = card_data['sign_image_ref']
        self.icon = card_data['icon_image_ref']
        self.puchi = f"https://hidamarirhodonite.kirara.ca/puchi/{self.card_id}.png"

        self.evo_id = card_data['evolution_id']
        self.evo_type = card_data['evolution_type']
        self.grow_type = card_data['grow_type']
        self.name = card_data['name']
        self.title = card_data['title']
        self.open_dress_id = card_data['open_dress_id']
        self.place = card_data['place']
        self.pose = card_data['pose']
        self.series_id = card_data['series_id']

        if card_data['skill'] is not None:
            self.skill = Skill(card_data['skill'])
        else:
            self.skill = None
        if card_data['lead_skill'] is not None:
            self.lead_skill = LeadSkill(card_data['lead_skill'])
        else:
            self.lead_skill = None

        self.rarity = card_data['rarity']

        self.min_vocal = card_data['vocal_min']
        self.max_vocal = card_data['vocal_max']
        self.bonus_dance = card_data['bonus_dance']

        self.min_dance = card_data['dance_min']
        self.max_dance = card_data['dance_max']
        self.bonus_dance = card_data['bonus_dance']

        self.min_visual = card_data['visual_min']
        self.max_visual = card_data['visual_max']
        self.bonus_visual = card_data['bonus_visual']

        self.min_hp = card_data['hp_min']
        self.max_hp = card_data['hp_max']
        self.bonus_hp = card_data['bonus_hp']

    def min_max_stats(self, stat: str, level: int):

        if stat == 'dance':
            dance_formula = self.min_dance + (self.max_dance - self.min_dance) * (level/self.rarity['base_max_level'])

            return round(dance_formula)

        elif stat == 'visual':
            visual_formula = self.min_visual + (self.max_visual - self.min_visual) * (level/self.rarity['base_max_level'])

            return round(visual_formula)

        elif stat == 'vocal':
            vocal_formula = self.min_vocal + (self.max_vocal - self.min_vocal) * (level/self.rarity['base_max_level'])

            return round(vocal_formula)


    


class Skill:
    def __init__(self, skill_data: dict):
        self.id = skill_data['id']
        self.name = skill_data['skill_name']
        self.explain = skill_data['explain']
        self.en_explain = skill_data['explain_en']

        self.skill_type = skill_data['skill_type']
        self.judge_type = skill_data['judge_type']
        self.trigger_type = skill_data['skill_trigger_type']
        self.trigger_value = skill_data['skill_trigger_value']

        self.cutin_type = skill_data['cutin_type']
        self.condition = skill_data['condition']

        self.value = skill_data['value']
        self.value_2 = skill_data['value_2']

        self.max_chance = skill_data['max_chance']
        self.max_duration = skill_data['max_duration']
        self.skill_type_id = skill_data['skill_type_id']
        self.effect_length = skill_data['effect_length']
        self.proc_chance = skill_data['proc_chance']


class LeadSkill:
    def __init__(self, skill_data: dict):
        self.id = skill_data['id']
        self.name = skill_data['name']
        self.explain = skill_data['explain']
        self.en_explain = skill_data['explain_en']
        self.type = skill_data['type']

        self.need_cute = bool(skill_data['need_cute'])
        self.need_cool = bool(skill_data['need_cool'])
        self.need_passion = bool(skill_data['need_passion'])

        self.target_attribute = skill_data['target_attribute']
        self.target_attribute_2 = skill_data['target_attribute_2']
        self.target_param = skill_data['target_param']
        self.target_param_2 = skill_data['target_param_2']

        self.up_type = skill_data['up_type']
        self.up_type_2 = skill_data['up_type_2']
        self.up_value = skill_data['up_value']
        self.up_value_2 = skill_data['up_value_2']

        self.special_id = skill_data['special_id']
        self.need_chara = skill_data['need_chara']



class Info:
    def __init__(self, info_data: dict):
        self.truth = info_data['truth_version']
        self._api_major = info_data['api_major']
        self._api_revision = info_data['api_revision']
        
class Gacha:
    def __init__(self, gacha_data: dict):
        self.id = gacha_data['id']
        self.name = gacha_data['name']
        self.start_date = gacha_data['start_date']
        self.end_date = gacha_data['end_date']
        self.type = gacha_data['type']
        self.subtype = gacha_data['subtype']
        self.rates = gacha_data['rates']

class Event:
    def __init__(self, gacha_data):
        self.id = gacha_data['id']
        self.name = gacha_data['name']
        self.start_date = gacha_data['start_date']
        self.end_date = gacha_data['end_date']
        self.result_end_date = gacha_data['result_end_date']